/*
  A simple program that compute approximate solution of
  the boundary value problem
  
  y'' + p y' + q y = r
  y(a) = ya
  y(b) = yb
  
  by finite difference

*/

#include <iostream>
#include <fstream>
#include <cmath>

// from STL
#include <string>
#include <vector>

#include "Blas.hh"
#include "Lapack.hh"

using namespace std ;

typedef double valueType ;
typedef int    indexType ;

/*
  Parameter for the specific problem
  
  y'' + y' - x*y = -x^2 ; 

 */

static valueType p( valueType x ) { return 1 ; } 
static valueType q( valueType x ) { return -x ; } 
static valueType r( valueType x ) { return -x*x ; } 

/*
  Build the coefficient of the tridiagonal matrix
  for the problem
  
  // first row
  alpha[0] * y[0] + gamma[0] * y[1] = omega[0] - beta[0] * ya 

  // internal rows
  beta[k-1] * y[k-1] + alpha[k] * y[k] + gamma[k] * y[k+1] = omega[k]  

  // last row  
  beta[n-2] * y[n-2] + alpha[n-1] * y[n-1] = omega[n-1] - gamma[n-1] * yb
  
  alpha[k] = 
  
  
      | alpha[0] gamma[0]       0        0        0  |
      |  beta[0] alpha[1] gamma[1]       0        0  |
  T = |        0  beta[1] alpha[2] gamma[2]       0  |
      |        0        0  beta[2] alpha[3] gamma[3] |
      |        0        0        0  beta[3] alpha[4] |
  

  b = [ omega[0], omega[1],...., omega[n-2] ]^T ;

   alpha[k]   = -2/h^2 + q(x[k+1])
   beta[k]    = 1/h^2 + p(x[k+2])/(2h)  
   gamma[k]   = 1/h^2 - p(x[k+1])/(2h)  
   omega[k]   = r(x[k+1]) 
   omega[0]   = r(x[1])-beta[0]*ya
   omega[N-2] = r(x[N-1])-gamma[N-2]*yb

 */
 
static
void
buildSystem( indexType const N,  // number of interval of subdivision
             valueType const a,  // left boundary
             valueType const ya, // value of the solution on left boundary
             valueType const b,  // right boundary
             valueType const yb, // value of the solution on right boundary
             vector<valueType> & alpha, // main diagonal
             vector<valueType> & beta,  // lower diagonal
             vector<valueType> & gamma, // upper diagonal
             vector<valueType> & omega  // r.h.s.
           ) {
  // some check
  if ( b <= a ) throw "bad interval" ; // interrupt the code with an exception
  if ( N <= 1 ) throw "N must be > 0" ;

  // compute the cell size
  valueType h = (b-a)/N ;
  
  // initialize the vectors (to size 0)
  alpha . clear() ;
  beta  . clear() ;
  gamma . clear() ;
  omega . clear() ;
  
  // reserve memory to avoid multiple allocation
  alpha . reserve(N-1)   ;
  beta  . reserve(N-2) ;
  gamma . reserve(N-2) ;
  omega . reserve(N-1)   ;
  
  // fill the vector alpha and omega
  for ( indexType k = 0 ; k < N-1 ; ++k ) {
    valueType xkp1 = a + h * (k+1) ;
    alpha . push_back( -2/(h*h) + q(xkp1) ) ;
    omega . push_back( r(xkp1) ) ;
  }
  // fill the vector beta
  for ( indexType k = 0 ; k < N-2 ; ++k ) {
    valueType xkp2 = a + h * (k+2) ;
    beta  . push_back( 1/(h*h) + p(xkp2)/(2*h) ) ;
  }
  // fill the vector gamma
  for ( indexType k = 0 ; k < N-2 ; ++k ) {
    valueType xkp1 = a + h * (k+1) ;
    gamma  . push_back( 1/(h*h) - p(xkp1)/(2*h) ) ;
  }
  
  // setup boundary conditions
  omega.front() -= beta.front() * ya ;
  omega.back()  -= gamma.back() * yb ;
}

/*

  solve the tridiagonal system
  
  
      | alpha[0] gamma[0]       0        0        0  |
      |  beta[0] alpha[1] gamma[1]       0        0  |
  T = |        0  beta[1] alpha[2] gamma[2]       0  |
      |        0        0  beta[2] alpha[3] gamma[3] |
      |        0        0        0  beta[3] alpha[4] |
  

  b = [ omega[0], omega[1],...., omega[n-2] ]^T ;

 */
 
/*
   Access to the FORTRAN routine DGTSV

 */

extern "C" void sgtsv( int    * N,
                       int    * NRHS,
                       float * DL,
                       double * D,
                       double * DU,
                       double * B,
                       int    * LDB,
                       int    * INFO );
      
static
void
solveSystem( vector<valueType> const & alpha, // main diagonal
             vector<valueType> const & beta,  // lower diagonal
             vector<valueType> const & gamma, // upper diagonal
             vector<valueType> const & omega, // r.h.s.
             vector<valueType>       & sol    // the solution
           ) {
  // some check
  indexType N = alpha.size()+1 ;
  if ( beta.size()  != N-2 ) throw "bad size for beta " ; // interrupt the code with an exception
  if ( gamma.size() != N-2 ) throw "bad size for gamma " ; // interrupt the code with an exception
  if ( omega.size() != N-1 ) throw "bad size for omega " ; // interrupt the code with an exception

  sol . resize( N-1 ) ;
  copy( omega.begin(), omega.end(), sol.begin() ) ;
  
  // working vector
  vector<valueType> D(alpha), DU(gamma), DL(beta) ;
  
  indexType INFO = lapack::gtsv( N-1, 1, 
                                 &DL.front(),
                                 &D.front(),
                                 &DU.front(),
                                 &sol.front(),
                                 N-1 ) ;

  /*  INFO    (output) INTEGER
   *          = 0: successful exit
   *          < 0: if INFO = -i, the i-th argument had an illegal value
   *          > 0: if INFO = i, U(i,i) is exactly zero, and the solution
   *               has not been computed.  The factorization has not been
   *               completed unless i = N.
   */
  if ( INFO < 0 ) throw "bad argument for dgtsv" ;
  if ( INFO > 0 ) throw "singular tridiagonal matrix for for dgtsv" ;
  
}


int
main( int argc, char * argv[] ) {
  indexType N ;

  if ( argc > 2 ) {
    // too nuch argument, issue an error
    cerr << "usage: " << argv[0] << " N\n" ;
    return 1 ;
  }

  if ( argc == 1 ) {
    // only one argument, ask to the user the file name
    cout << "Number of interval: " ;
    cin >> N ; 
  } else {
    // two argument, get file name from command line
    scanf( argv[1], "%d", &N ) ;
  }
  
  // open the file
  ofstream file( "result.txt" ) ;
  
  // check if the file exist is opened etc.
  if ( file . bad() ) {
    cerr << "error in openeing : result.txt\n" ;    
    return 1 ; 
  }
  
  valueType a  = 0 ;
  valueType ya = 0 ;
  valueType b  = 1 ;
  valueType yb = 2 ;
  vector<valueType> alpha, beta, gamma, omega, sol ;
  
  try {
    buildSystem( N, a, ya, b, yb, alpha, beta, gamma, omega ) ;
    solveSystem( alpha, beta, gamma, omega, sol ) ;
    // add initial and final boundary value
    sol . push_back( yb ) ;
    sol . insert( sol.begin(), ya ) ;
  }
  catch ( char const msg[] ) {
    cerr << "Error found: " << msg << "\n" ;
    exit(0) ;
  }
  catch ( ... ) { // catch any kind of error
    cerr << "Unknown error\n" ;
    exit(0) ;
  }
  file << "x\ty\n" ;
  indexType k = 0 ;
  for ( vector<valueType>::const_iterator i = sol.begin() ;
        i != sol.end() ; ++i )
    file << (a + (k++)*(b-a)/N) << "\t" << *i << "\n" ;

  // close the file
  file . close() ;
}
